@@ -1,3 +1,4 @@
+import datetime
 import re
 
 COMPILED_REGEX_TYPE = type(re.compile(''))
@@ -10,3 +11,7 @@ class RegexObject(object):
 
     def __eq__(self, other):
         return self.pattern == other.pattern and self.flags == other.flags
+
+
+def get_migration_name_timestamp():
+    return datetime.datetime.now().strftime("%Y%m%d_%H%M")
